/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Jun Furuse, projet Cristal, INRIA Rocquencourt           */
/*                                                                     */
/*  Copyright 1999,2000                                                */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  Distributed only by permission.                                    */
/*                                                                     */
/***********************************************************************/
#include <config.h>
#include <string.h>

#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/callback.h>
#include <caml/fail.h>

#if (HAVE_FREETYPE && HAVE_FREETYPE2)

#include <freetype/freetype.h>

value init_FreeType()
{
  CAMLparam0();
  FT_Library *library;

  if( (library = stat_alloc( sizeof(FT_Library) )) == NULL ){
    failwith( "Memory over" );
  }
  if( FT_Init_FreeType( library ) ){
    failwith( "FT_Init_FreeType failed" );
  }
  CAMLreturn( (value) library );
}

value done_FreeType( library )
     value library;
{
  CAMLparam1(library);
  FT_Done_FreeType( *(FT_Library *)library );
  stat_free( (void *) library );
  CAMLreturn(Val_unit);
}

#include <stdio.h>
value new_Face( library, fontpath, idx )
     value library;
     value fontpath;
     value idx;
{
  CAMLparam3(library, fontpath, idx );
  FT_Face *face;
  
  if( (face = stat_alloc( sizeof(FT_Face) )) == NULL ){
    failwith( "Memory over" );
  }
  if( FT_New_Face( *(FT_Library *)library, String_val( fontpath ), Int_val( idx ), face ) ){
    failwith( "Could not open face" );
  }
  CAMLreturn( (value) face );
}

value face_info( facev )
     value facev;
{
  CAMLparam1(facev);
  CAMLlocal1(res);

  FT_Face face = *(FT_Face *)facev;
  res = alloc_tuple(14);
  Field(res, 0) = Val_int( face->num_faces );
  Field(res, 1) = Val_int( face->num_glyphs );
  Field(res, 2) = copy_string( face->family_name == NULL ? "" : face->family_name );
  Field(res, 3) = copy_string( face->style_name == NULL ? "" : face->style_name );
  Field(res, 4) = Val_bool( FT_HAS_HORIZONTAL( face ) );
  Field(res, 5) = Val_bool( FT_HAS_VERTICAL( face ) );
  Field(res, 6) = Val_bool( FT_HAS_KERNING( face ) );
  Field(res, 7) = Val_bool( FT_IS_SCALABLE( face ) );
  Field(res, 8) = Val_bool( FT_IS_SFNT( face ) );
  Field(res, 9) = Val_bool( FT_IS_FIXED_WIDTH( face ) );
  Field(res,10) = Val_bool( FT_HAS_FIXED_SIZES( face ) );
  Field(res,11) = Val_bool( FT_HAS_FAST_GLYPHS( face ) );
  Field(res,12) = Val_bool( FT_HAS_GLYPH_NAMES( face ) );
  Field(res,13) = Val_bool( FT_HAS_MULTIPLE_MASTERS( face ) );
  
  CAMLreturn(res);
}

value done_Face( face )
     value face;
{
  CAMLparam1(face);
  FT_Done_Face( *(FT_Face *) face );
  CAMLreturn( Val_unit );
}

value get_num_glyphs( face )
     value face;
{
  CAMLparam1(face);
  CAMLreturn( Val_int ((*(FT_Face *) face)->num_glyphs) );
}


value set_Char_Size( face, char_w, char_h, res_h, res_v )
     value face;
     value char_w, char_h; /* 26.6 1 = 1/64pt */
     value res_h, res_v; /* dpi */
{
  CAMLparam5( face, char_w, char_h, res_h, res_v );
  if ( FT_Set_Char_Size( *(FT_Face *) face,
		    Int_val(char_w), Int_val(char_h),
			 Int_val(res_h), Int_val(res_v) ) ){
    failwith("set char size failed");
  }
  CAMLreturn(Val_unit);
}

/* to be done: query at face->fixed_sizes
 */

value set_Pixel_Sizes( face, pixel_w, pixel_h )
     value face;
     value pixel_w, pixel_h; /* dot */
{
  CAMLparam3(face,pixel_w,pixel_h);
  if ( FT_Set_Pixel_Sizes( *(FT_Face *) face,
			 Int_val(pixel_w), Int_val(pixel_h) ) ){
    failwith("set pixel sizes failed");
  }
  CAMLreturn(Val_unit);
}

value val_CharMap( charmapp )
     FT_CharMap *charmapp;
{
  CAMLparam0();
  CAMLlocal1(res);
  
  res = alloc_tuple(2);
  Field(res,0) = Val_int((*charmapp)->platform_id);
  Field(res,1) = Val_int((*charmapp)->encoding_id);

  CAMLreturn(res);
}

value get_CharMaps( facev )
     value facev;
{
  CAMLparam1(facev);
  CAMLlocal3(list,last_cell,new_cell);
  int i = 0;
  FT_Face face;

  face = *(FT_Face *) facev;

  list = last_cell = Val_unit;
  
  while( i < face->num_charmaps ){
    new_cell = alloc_tuple(2);
    Field(new_cell,0) = val_CharMap( face->charmaps + i );
    Field(new_cell,1) = Val_unit;
    if( i == 0 ){
      list = new_cell;
    } else {
      modify(&Field(last_cell,1), new_cell);
    }
    last_cell = new_cell;
    i++;
  }

  CAMLreturn(list);
}

value set_CharMap( facev, charmapv )
     value facev;
     value charmapv;
{
  CAMLparam2(facev,charmapv);
  int i = 0;
  FT_Face face;
  FT_CharMap charmap;
  int my_pid, my_eid;
  
  face = *(FT_Face *) facev;
  my_pid = Int_val(Field(charmapv, 0));
  my_eid = Int_val(Field(charmapv, 1));

  while( i < face->num_charmaps ){
    charmap = face->charmaps[i];
    if ( charmap->platform_id == my_pid && 
	 charmap->encoding_id == my_eid ){
      FT_Set_Charmap( face, charmap );
      CAMLreturn(Val_unit);
    } else {
      i++;
    }
  }
  failwith("freetype:set_charmaps: selected pid+eid do not exist");
}

value get_Char_Index( face, code )
     value face, code;
{
  CAMLparam2(face,code);
  CAMLreturn(Val_int(FT_Get_Char_Index( *(FT_Face *)face, Int_val(code))));
}

value load_Glyph( face, index, flags )
     value face, index, flags;
{
  CAMLparam3(face,index,flags);
  CAMLlocal1(res);

  FT_Load_Glyph( *(FT_Face *) face, Int_val(index), FT_LOAD_DEFAULT | Int_val(flags)); 

  res = alloc_tuple(2);
  Field(res,0) = Val_int( (*(FT_Face*)face)->glyph->advance.x );
  Field(res,1) = Val_int( (*(FT_Face*)face)->glyph->advance.y );

  CAMLreturn(res);
}

value load_Char( face, code, flags )
     value face, code, flags;
{
  CAMLparam3(face,code,flags);
  CAMLlocal1(res);

  /* FT_Load_Glyph(face, FT_Get_Char_Index( face, code )) */
  FT_Load_Char( *(FT_Face *) face, Int_val(code), FT_LOAD_DEFAULT | Int_val(flags)); 

  res = alloc_tuple(2);
  Field(res,0) = Val_int( (*(FT_Face*)face)->glyph->advance.x );
  Field(res,1) = Val_int( (*(FT_Face*)face)->glyph->advance.y );

  CAMLreturn(res);
}

value render_Glyph_of_Face( face )
     value face;
{
  CAMLparam1(face);
  if( FT_Render_Glyph( (*(FT_Face *)face)->glyph , ft_render_mode_normal ) ){
    failwith("render_Glyph_of_Face() failed");
  }
  CAMLreturn(Val_unit);
}

value render_Char( face, code, flags )
     value face, code, flags;
{
  CAMLparam3(face,code,flags);
  CAMLlocal1(res);

  /* FT_Load_Glyph(face, FT_Get_Char_Index( face, code ), FT_LOAD_RENDER) */
  if( FT_Load_Char( *(FT_Face *) face, Int_val(code), FT_LOAD_RENDER | Int_val(flags))) { 
    failwith("render_Char() failed");
  }

  res = alloc_tuple(2);
  Field(res,0) = Val_int( (*(FT_Face*)face)->glyph->advance.x );
  Field(res,1) = Val_int( (*(FT_Face*)face)->glyph->advance.y );

  CAMLreturn(res);
}

value set_Transform( face, vmatrix, vpen ) 
     value face, vmatrix, vpen;
{
  CAMLparam3(face, vmatrix, vpen);
  FT_Matrix matrix;
  FT_Vector pen;

  matrix.xx = (FT_Fixed)( Int_val(Field(vmatrix,0)) );
  matrix.xy = (FT_Fixed)( Int_val(Field(vmatrix,1)) );
  matrix.yx = (FT_Fixed)( Int_val(Field(vmatrix,2)) );
  matrix.yy = (FT_Fixed)( Int_val(Field(vmatrix,3)) );
  pen.x = (FT_Fixed)( Int_val(Field(vpen,0)) );
  pen.y = (FT_Fixed)( Int_val(Field(vpen,1)) );

  FT_Set_Transform( *(FT_Face *)face, &matrix, &pen );
  
  CAMLreturn(Val_unit);
}

value get_Bitmap_Info( vface )
     value vface;
{
  CAMLparam1(vface);
  CAMLlocal1(res);

  FT_GlyphSlot glyph = (*(FT_Face *)vface)->glyph;
  FT_Bitmap bitmap = glyph->bitmap;

  if ( bitmap.pixel_mode != ft_pixel_mode_grays ){
    fprintf(stderr, "pixel_mode=%d\n", bitmap.pixel_mode); 
    failwith("unknown pixel mode");
  }
  if ( bitmap.num_grays != 256 ){
    failwith("unknown num_grays");
  }
  res = alloc_tuple(4);
  Field(res,0) = Val_int(glyph->bitmap_left);
  Field(res,1) = Val_int(glyph->bitmap_top);
  Field(res,2) = Val_int(bitmap.width);
  Field(res,3) = Val_int(bitmap.rows);

  CAMLreturn(res);
}

value read_Bitmap( vface, vx, vy ) /* This "y" is in Y upwards convention */
     value vface, vx, vy;
{
  /* no boundary check !!! */
  int x, y;
  CAMLparam3(vface, vx, vy);

  FT_Bitmap bitmap = (*(FT_Face *)vface)->glyph->bitmap;

  unsigned char *row;

  x = Int_val(vx);
  y = Int_val(vy);
  
  if (bitmap.pitch > 0){
    row = bitmap.buffer + (bitmap.rows - 1 - y) * bitmap.pitch;
  } else {
    row = bitmap.buffer - y * bitmap.pitch;
  }
  CAMLreturn (Val_int(row[x]));
}

value get_Glyph_Metrics( face )
     value face;
{
  CAMLparam1(face);
  CAMLlocal3(res1,res2,res);
  
  /* no soundness check ! */
  FT_Glyph_Metrics *metrics = &((*(FT_Face *)face)->glyph->metrics);
 
  res1 = alloc_tuple(3);  
  Field(res1,0) = Val_int(metrics->horiBearingX);
  Field(res1,1) = Val_int(metrics->horiBearingY);
  Field(res1,2) = Val_int(metrics->horiAdvance);
			  
  res2 = alloc_tuple(3);  
  Field(res2,0) = Val_int(metrics->vertBearingX);
  Field(res2,1) = Val_int(metrics->vertBearingY);
  Field(res2,2) = Val_int(metrics->vertAdvance);

  res = alloc_tuple(3);
  Field(res,0) = Val_int(metrics->width);
  Field(res,1) = Val_int(metrics->height);
  Field(res,2) = res1;
  Field(res,3) = res2;

  CAMLreturn(res);
}

value get_Size_Metrics( face )
     value face;
{
  CAMLparam1(face);
  CAMLlocal1(res);

  FT_Size_Metrics *imetrics = &((*(FT_Face*)face)->size->metrics);

  res = alloc_tuple(4);
  Field(res,0) = Val_int(imetrics->x_ppem);
  Field(res,1) = Val_int(imetrics->y_ppem);
  Field(res,2) = Val_int(imetrics->x_scale);
  Field(res,3) = Val_int(imetrics->y_scale);

  CAMLreturn(res);
}

value get_Outline_Contents(value face) {
/* *****************************************************************
    
   Concrete definitions of TT_Outline might vary from version to
   version.

   This definition assumes freetype 2.0.1
     
 ***************************************************************** */
  CAMLparam1(face);
  CAMLlocal5(points, tags, contours, res, tmp);
  int i;

  FT_Outline* outline = &((*(FT_Face *)face)->glyph->outline);

  int n_contours = outline->n_contours;
  int n_points   = outline->n_points;

  points   = alloc_tuple(n_points);
  tags     = alloc_tuple(n_points);
  contours = alloc_tuple(n_contours);

  for( i=0; i<n_points; i++ ) {
    FT_Vector* raw_points = outline->points;
    char* raw_flags  = outline->tags;
    tmp = alloc_tuple(2);
    /* caution: 26.6 fixed into 31 bit */
    Store_field(tmp, 0, Val_int(raw_points[i].x)); 
    Store_field(tmp, 1, Val_int(raw_points[i].y)); 
    Store_field(points, i, tmp);
    Store_field(tags, i, Val_int(raw_flags[i] & 0x01));
  }

  for( i=0; i<n_contours; i++ ) {
    short* raw_contours = outline->contours;
    Store_field(contours, i, Val_int(raw_contours[i]));
  }
  
  res = alloc_tuple(5);
  Store_field(res, 0, Val_int(n_contours));
  Store_field(res, 1, Val_int(n_points));
  Store_field(res, 2, points);
  Store_field(res, 3, tags);
  Store_field(res, 4, contours);
  
  CAMLreturn(res);
}

#endif
